from pydmoo.algorithms.modern.moeadde_imkt_clstm import (
    MOEADDEIMicLSTM1003,
    MOEADDEIMicLSTM1005,
    MOEADDEIMicLSTM1007,
    MOEADDEIMicLSTM1009,
    MOEADDEIMicLSTM1503,
    MOEADDEIMicLSTM1505,
    MOEADDEIMicLSTM1507,
    MOEADDEIMicLSTM1509,
    MOEADDEIMicLSTM1511,
    MOEADDEIMicLSTM1513,
)
from pydmoo.algorithms.modern.moeadde_imkt_lstm import (
    MOEADDEIMiLSTM1003,
    MOEADDEIMiLSTM1005,
    MOEADDEIMiLSTM1007,
    MOEADDEIMiLSTM1009,
    MOEADDEIMiLSTM1503,
    MOEADDEIMiLSTM1505,
    MOEADDEIMiLSTM1507,
    MOEADDEIMiLSTM1509,
    MOEADDEIMiLSTM1511,
    MOEADDEIMiLSTM1513,
)
from pydmoo.algorithms.modern.nsga2_imkt_clstm import (
    NSGA2IMicLSTM1003,
    NSGA2IMicLSTM1005,
    NSGA2IMicLSTM1007,
    NSGA2IMicLSTM1009,
    NSGA2IMicLSTM1503,
    NSGA2IMicLSTM1505,
    NSGA2IMicLSTM1507,
    NSGA2IMicLSTM1509,
    NSGA2IMicLSTM1511,
    NSGA2IMicLSTM1513,
)
from pydmoo.algorithms.modern.nsga2_imkt_lstm import (
    NSGA2IMiLSTM1003,
    NSGA2IMiLSTM1005,
    NSGA2IMiLSTM1007,
    NSGA2IMiLSTM1009,
    NSGA2IMiLSTM1503,
    NSGA2IMiLSTM1505,
    NSGA2IMiLSTM1507,
    NSGA2IMiLSTM1509,
    NSGA2IMiLSTM1511,
    NSGA2IMiLSTM1513,
)

ALGORITHMS_ABLATION = {
    "MOEADDEIMicLSTM1003": MOEADDEIMicLSTM1003,
    "MOEADDEIMicLSTM1005": MOEADDEIMicLSTM1005,
    "MOEADDEIMicLSTM1007": MOEADDEIMicLSTM1007,
    "MOEADDEIMicLSTM1009": MOEADDEIMicLSTM1009,
    "MOEADDEIMicLSTM1503": MOEADDEIMicLSTM1503,
    "MOEADDEIMicLSTM1505": MOEADDEIMicLSTM1505,
    "MOEADDEIMicLSTM1507": MOEADDEIMicLSTM1507,
    "MOEADDEIMicLSTM1509": MOEADDEIMicLSTM1509,
    "MOEADDEIMicLSTM1511": MOEADDEIMicLSTM1511,
    "MOEADDEIMicLSTM1513": MOEADDEIMicLSTM1513,

    "MOEADDEIMiLSTM1003": MOEADDEIMiLSTM1003,
    "MOEADDEIMiLSTM1005": MOEADDEIMiLSTM1005,
    "MOEADDEIMiLSTM1007": MOEADDEIMiLSTM1007,
    "MOEADDEIMiLSTM1009": MOEADDEIMiLSTM1009,
    "MOEADDEIMiLSTM1503": MOEADDEIMiLSTM1503,
    "MOEADDEIMiLSTM1505": MOEADDEIMiLSTM1505,
    "MOEADDEIMiLSTM1507": MOEADDEIMiLSTM1507,
    "MOEADDEIMiLSTM1509": MOEADDEIMiLSTM1509,
    "MOEADDEIMiLSTM1511": MOEADDEIMiLSTM1511,
    "MOEADDEIMiLSTM1513": MOEADDEIMiLSTM1513,

    "NSGA2IMicLSTM1003": NSGA2IMicLSTM1003,
    "NSGA2IMicLSTM1005": NSGA2IMicLSTM1005,
    "NSGA2IMicLSTM1007": NSGA2IMicLSTM1007,
    "NSGA2IMicLSTM1009": NSGA2IMicLSTM1009,
    "NSGA2IMicLSTM1503": NSGA2IMicLSTM1503,
    "NSGA2IMicLSTM1505": NSGA2IMicLSTM1505,
    "NSGA2IMicLSTM1507": NSGA2IMicLSTM1507,
    "NSGA2IMicLSTM1509": NSGA2IMicLSTM1509,
    "NSGA2IMicLSTM1511": NSGA2IMicLSTM1511,
    "NSGA2IMicLSTM1513": NSGA2IMicLSTM1513,

    "NSGA2IMiLSTM1003": NSGA2IMiLSTM1003,
    "NSGA2IMiLSTM1005": NSGA2IMiLSTM1005,
    "NSGA2IMiLSTM1007": NSGA2IMiLSTM1007,
    "NSGA2IMiLSTM1009": NSGA2IMiLSTM1009,
    "NSGA2IMiLSTM1503": NSGA2IMiLSTM1503,
    "NSGA2IMiLSTM1505": NSGA2IMiLSTM1505,
    "NSGA2IMiLSTM1507": NSGA2IMiLSTM1507,
    "NSGA2IMiLSTM1509": NSGA2IMiLSTM1509,
    "NSGA2IMiLSTM1511": NSGA2IMiLSTM1511,
    "NSGA2IMiLSTM1513": NSGA2IMiLSTM1513,
}
